# -*- encoding: utf-8 -*-
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# @brief Code generated by Mobius. Do not edit
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
EXTENSION_ID = 'hive-report'
EXTENSION_NAME = 'Hive Report'
EXTENSION_AUTHOR = 'Eduardo Aguiar and Vladimir Santos'
EXTENSION_VERSION = '0.1.4'
EXTENSION_DESCRIPTION = 'Add reports to Hive extension'

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# Mobius Forensic Toolkit
# Copyright (C) 2008,2009,2010,2011 Eduardo Aguiar
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
import struct
import datetime
import re
import hashlib
import hmac
import binascii
import urllib
import urlparse

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# @brief get NT datetime from timestamp
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
def get_nt_datetime (timestamp):
  if timestamp:
    delta = datetime.timedelta (seconds=timestamp // 10000000)
    date = datetime.datetime (1601, 1, 1, 0, 0, 0) + delta
    text = date.strftime ('%F %T')
  else:
    text = ''

  return text

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# @brief get Unix datetime from registry value
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
def get_unix_datetime (timestamp):
  if timestamp:
    delta = datetime.timedelta (seconds=timestamp)
    date = datetime.datetime (1970, 1, 1, 0, 0, 0) + delta
    text = date.strftime ('%F %T')
  else:
    text = ''

  return text

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# @brief get username from SID
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
def get_username_from_sid (registry, sid):
  machine_sid = registry.get_data ('HKLM/SECURITY/Policy/PolAcDMS/(default)') or ''

  if machine_sid != '':
    machine_sid = '-'.join ([str(i) for i in struct.unpack ('<cB6xBIIII',
                            'S' + machine_sid)])

  username = None

  if sid.startswith (machine_sid):
    rid = int (sid.split ('-')[-1])
    data = registry.get_data ('HKLM/SAM/SAM/Domains/Account/Users/%08X/V' % rid)

    if data is not None:
      offset, size = struct.unpack ('<II', data[12:20])
      username = data[0xcc + offset:0xcc + offset + size].decode ('utf-16-le')

  return username

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# @brief expand EXPAND_SZ
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
def expand_sz (value, environment_vars):

  # replace function
  def repl (matchobj):
    name = matchobj.group (1).lower ()
    return environment_vars.get (name, '%%%s%%' % name)

  return re.sub (r'%(.*?)%', repl, value or '')

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# @brief get product key from digital product ID
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
def get_product_key (digital_product_id):
  if digital_product_id is None:
    return None

  digits = 'BCDFGHJKMPQRTVWXY2346789'
  data = list (digital_product_id[52:68])
  out = ''
    
  for i in range (29):
    if (i + 1) % 6 == 0:
      out = '-' + out
    else:
      idx = 0
      for j in range (14, -1, -1):
        b = (idx << 8) | ord (data[j])
        data[j] = chr (b // 24)
        idx = b % 24
      out = digits[idx] + out
        
  return out

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# @brief Generates candidate passwords
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
def candidate_passwords(reg):
  values = []

  for item in list_lsa_secrets(reg):
    values.extend([x for x in item[1:3] if x != ''])

  for item in list_pstore(reg):
    if item[0] is not None:
      values.append(item[0])

  passwords = set([''])
  for value in values:
    for encoding in ('utf-16-le', 'utf-8'):
      try:
        passwords.update(value.decode(encoding).rstrip('\x00').split('\x00'))
      except Exception as e:
        pass

  return passwords

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# RC4 cipher algorithm
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
class RC4(object):
  '''Alleged RC4 cipher

  Based on the posting 'RC4 Algorithm revealed' to Usenet
  newsgroups sci.crypt, alt.security, comp.security.misc and
  alt.privacy on September 14, 1994.

  http://en.wikipedia.org/wiki/RC4
  http://www.mozilla.org/projects/security/pki/nss/draft-kaukonen-cipher-arcfour-03.txt
  '''
  def __init__(self, key):
    '''Creates a new RC4 object.
    '''
    if len(key) < 1 or len(key) > 256:
      raise ValueError('Key size must be between 1 and 256 bytes')

    self.__key = [ord(c) for c in key]
    self._i, self._j, self._S = self.__keyschedule(self.__key)

  def __keyschedule(self, key):
    '''Key-scheduling algorithm (KSA).
    '''
    j, S = 0, range(256)
    for i in range(256):
      j = (j + S[i] + key[i % len(key)]) % 256
      S[i], S[j] = S[j], S[i]
    i, j = 0, 0

    return i, j, S

  def __nextbyte(self, nbytes = 1):
    '''Pseudo-random generation algorithm (PRGA).
    '''
    i, j, S = self._i, self._j, self._S

    result = []
    for n in range(nbytes):
      i = (i + 1) % 256
      j = (j + S[i]) % 256
      S[i], S[j] = S[j], S[i]
      result.append(S[(S[i] + S[j]) % 256])

    self._i, self._j, self._S = i, j, S

    return result

  def __crypt(self, message):
    '''Encrypts or decrypts a message.
    '''
    return ''.join([chr(b ^ ord(c)) for b, c in
                    zip(self.__nextbyte(len(message)), message)])

  def copy(self):
    '''Makes a separate copy of this cipher object.
    '''
    other = self.__class__(self.key)
    other._i, other._j, other._S = self._i, self._j, self._S[:]
    return other

  def encrypt(self, message):
    '''Encrypts a message.
    '''
    return self.__crypt(message)
 
  def decrypt(self, message):
    '''Decrypts a message.
    '''
    return self.__crypt(message)

  @property
  def name(self):
    '''Name of the cipher algorithm.
    '''
    return 'rc4'

  @property
  def key(self):
    '''Current key of the cipher.
    '''
    return ''.join([chr(b) for b in self.__key])

  @property
  def block_size(self):
    '''Current block size of the cipher in bytes.
    '''
    return 1

  @property
  def key_size(self):
    '''Current key size of the cipher in bytes.
    '''
    return len(self.__key)

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# # DES cipher algorithm
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
class DES(object):
  '''DES cipher

  Based on the Data Encryption Standard defined by the National
  Institute of Standards and Technology (NIST) of the U.S. Department
  of Commerce in the Federal Information Processing Standards
  Publication 46-3 (FIPS PUB 46-3).
  '''
  # Initial permutation (IP)
  __IP = (
    57, 49, 41, 33, 25, 17,  9,  1,
    59, 51, 43, 35, 27, 19, 11,  3,
    61, 53, 45, 37, 29, 21, 13,  5,
    63, 55, 47, 39, 31, 23, 15,  7,
    56, 48, 40, 32, 24, 16,  8,  0,
    58, 50, 42, 34, 26, 18, 10,  2,
    60, 52, 44, 36, 28, 20, 12,  4,
    62, 54, 46, 38, 30, 22, 14,  6
  )
  # Inverse of the initial permutation (IP^-1)
  __IP_1 = (
    39,  7, 47, 15, 55, 23, 63, 31,
    38,  6, 46, 14, 54, 22, 62, 30,
    37,  5, 45, 13, 53, 21, 61, 29,
    36,  4, 44, 12, 52, 20, 60, 28,
    35,  3, 43, 11, 51, 19, 59, 27,
    34,  2, 42, 10, 50, 18, 58, 26,
    33,  1, 41,  9, 49, 17, 57, 25,
    32,  0, 40,  8, 48, 16, 56, 24
  )
  # E bit-selection table
  __E = (
    31,  0,  1,  2,  3,  4,
     3,  4,  5,  6,  7,  8,
     7,  8,  9, 10, 11, 12,
    11, 12, 13, 14, 15, 16,
    15, 16, 17, 18, 19, 20,
    19, 20, 21, 22, 23, 24,
    23, 24, 25, 26, 27, 28,
    27, 28, 29, 30, 31,  0
  )
  # Selection functions (S1 to S8)
  __S = (
    (
       14,  4, 13,  1,  2, 15, 11,  8,  3, 10,  6, 12,  5,  9,  0,  7,
        0, 15,  7,  4, 14,  2, 13,  1, 10,  6, 12, 11,  9,  5,  3,  8,
        4,  1, 14,  8, 13,  6,  2, 11, 15, 12,  9,  7,  3, 10,  5,  0,
       15, 12,  8,  2,  4,  9,  1,  7,  5, 11,  3, 14, 10,  0,  6, 13
    ),
    (
       15,  1,  8, 14,  6, 11,  3,  4,  9,  7,  2, 13, 12,  0,  5, 10,
        3, 13,  4,  7, 15,  2,  8, 14, 12,  0,  1, 10,  6,  9, 11,  5,
        0, 14,  7, 11, 10,  4, 13,  1,  5,  8, 12,  6,  9,  3,  2, 15,
       13,  8, 10,  1,  3, 15,  4,  2, 11,  6,  7, 12,  0,  5, 14,  9
    ),
    (
       10,  0,  9, 14,  6,  3, 15,  5,  1, 13, 12,  7, 11,  4,  2,  8,
       13,  7,  0,  9,  3,  4,  6, 10,  2,  8,  5, 14, 12, 11, 15,  1,
       13,  6,  4,  9,  8, 15,  3,  0, 11,  1,  2, 12,  5, 10, 14,  7,
        1, 10, 13,  0,  6,  9,  8,  7,  4, 15, 14,  3, 11,  5,  2, 12
    ),
    (
        7, 13, 14,  3,  0,  6,  9, 10,  1,  2,  8,  5, 11, 12,  4, 15,
       13,  8, 11,  5,  6, 15,  0,  3,  4,  7,  2, 12,  1, 10, 14,  9,
       10,  6,  9,  0, 12, 11,  7, 13, 15,  1,  3, 14,  5,  2,  8,  4,
        3, 15,  0,  6, 10,  1, 13,  8,  9,  4,  5, 11, 12,  7,  2, 14
    ),
    (
        2, 12,  4,  1,  7, 10, 11,  6,  8,  5,  3, 15, 13,  0, 14,  9,
       14, 11,  2, 12,  4,  7, 13,  1,  5,  0, 15, 10,  3,  9,  8,  6,
        4,  2,  1, 11, 10, 13,  7,  8, 15,  9, 12,  5,  6,  3,  0, 14,
       11,  8, 12,  7,  1, 14,  2, 13,  6, 15,  0,  9, 10,  4,  5,  3
    ),
    (
       12,  1, 10, 15,  9,  2,  6,  8,  0, 13,  3,  4, 14,  7,  5, 11,
       10, 15,  4,  2,  7, 12,  9,  5,  6,  1, 13, 14,  0, 11,  3,  8,
        9, 14, 15,  5,  2,  8, 12,  3,  7,  0,  4, 10,  1, 13, 11,  6,
        4,  3,  2, 12,  9,  5, 15, 10, 11, 14,  1,  7,  6,  0,  8, 13
    ),
    (
        4, 11,  2, 14, 15,  0,  8, 13,  3, 12,  9,  7,  5, 10,  6,  1,
       13,  0, 11,  7,  4,  9,  1, 10, 14,  3,  5, 12,  2, 15,  8,  6,
        1,  4, 11, 13, 12,  3,  7, 14, 10, 15,  6,  8,  0,  5,  9,  2,
        6, 11, 13,  8,  1,  4, 10,  7,  9,  5,  0, 15, 14,  2,  3, 12
    ),
    (
       13,  2,  8,  4,  6, 15, 11,  1, 10,  9,  3, 14,  5,  0, 12,  7,
        1, 15, 13,  8, 10,  3,  7,  4, 12,  5,  6, 11,  0, 14,  9,  2,
        7, 11,  4,  1,  9, 12, 14,  2,  0,  6, 10, 13, 15,  3,  5,  8,
        2,  1, 14,  7,  4, 10,  8, 13, 15, 12,  9,  0,  3,  5,  6, 11
    )
  )
  # Permutation function
  __P = (
    15,  6, 19, 20,
    28, 11, 27, 16,
     0, 14, 22, 25,
     4, 17, 30,  9,
     1,  7, 23, 13,
    31, 26,  2,  8,
    18, 12, 29,  5,
    21, 10,  3, 24
  )
  # Permuted choice 1
  __PC1 = (
    56, 48, 40, 32, 24, 16,  8,
     0, 57, 49, 41, 33, 25, 17,
     9,  1, 58, 50, 42, 34, 26,
    18, 10,  2, 59, 51, 43, 35,
    62, 54, 46, 38, 30, 22, 14,
     6, 61, 53, 45, 37, 29, 21,
    13,  5, 60, 52, 44, 36, 28,
    20, 12,  4, 27, 19, 11,  3
  )
  # Schedule of left shifts (rotations)
  __LSHIFT = (
    1, 1, 2, 2, 2, 2, 2, 2, 1, 2, 2, 2, 2, 2, 2, 1
  )
  # Permuted choice 2
  __PC2 = (
    13, 16, 10, 23,  0,  4,
     2, 27, 14,  5, 20,  9,
    22, 18, 11,  3, 25,  7,
    15,  6, 26, 19, 12,  1,
    40, 51, 30, 36, 46, 54,
    29, 39, 50, 44, 32, 47,
    43, 48, 38, 55, 33, 52,
    45, 41, 49, 35, 28, 31
  )

  def __init__(self, key, mode='ecb', iv='\x00\x00\x00\x00\x00\x00\x00\x00'):
    '''Creates a new DES object.
    '''
    mode = mode.lower()
    if mode not in ('ecb', 'cbc'):
      raise ValueError('Mode must be ECB or CBC')
    self.__mode = mode

    if mode == 'cbc' and len(iv) != self.block_size:
      raise ValueError('Initialization vector size must be 8 bytes')
    self.__iv = [ord(c) for c in iv]

    self.__key = self.__fixparity(key)
    self.__kn = self.__keyschedule(self.__key)

  def __fixparity(self, key):
    '''Sets the last bit in every byte in key to reflect the odd
    parity. If a seven byte key is used, it will be expanded into
    eight bytes.
    '''
    if len(key) not in (7, 8):
      raise ValueError('Key size must be 7 or 8 bytes')

    key = [ord(c) for c in key]

    # expand seven byte key into eight bytes
    if len(key) == 7:
      key = [key[0] & 0xfe,
             ((key[0] & 0x01) << 7) | ((key[1] & 0xfc) >> 1),
             ((key[1] & 0x03) << 6) | ((key[2] & 0xf8) >> 2),
             ((key[2] & 0x07) << 5) | ((key[3] & 0xf0) >> 3),
             ((key[3] & 0x0f) << 4) | ((key[4] & 0xe0) >> 4),
             ((key[4] & 0x1f) << 3) | ((key[5] & 0xc0) >> 5),
             ((key[5] & 0x3f) << 2) | ((key[6] & 0x80) >> 6),
             (key[6] & 0x7f) << 1]

    # set odd parity for each byte of the key
    for i in range(self.key_size):
      parity = [1 for m in [1, 2, 4, 8, 16, 32, 64, 128] if (key[i] & m) > 0]
      if (len(parity) & 0x01) == 0:
        key[i] = (key[i] & 0xfe) | ((key[i] & 0x01) ^ 0x01)

    return key

  def __keyschedule(self, key):
    '''Key scheduling.
    '''
    kn = [0] * 16
    pc = self.__permutate(DES.__PC1, key)
    for n in range(16):
      # rotate left each half
      for i in range(DES.__LSHIFT[n]):
        msb = [(b & 0x80) >> 7 for b in pc[1:]]
        msb.append((pc[3] & 0x08) >> 3)
        msb.append((pc[0] & 0x80) >> 3)
        pc = [((pc[i] << 1) & 0xff) | msb[i] for i in range(7)]
        pc[3] = (pc[3] & 0xef) | msb[7]
      kn[n] = self.__permutate(DES.__PC2, pc)

    return kn

  def __permutate(self, table, data):
    '''Permutates data bits accordingly to the specified table.
    '''
    bytenum = len(table) >> 3
    x = [0] * bytenum
    for i in range(bytenum):
      tt = table[i << 3:(i << 3) + 8]
      x[i] = (((data[tt[0] >> 3] << (tt[0] & 7)) & 0x80) |
              (((data[tt[1] >> 3] << (tt[1] & 7)) & 0x80) >> 1) |
              (((data[tt[2] >> 3] << (tt[2] & 7)) & 0x80) >> 2) |
              (((data[tt[3] >> 3] << (tt[3] & 7)) & 0x80) >> 3) |
              (((data[tt[4] >> 3] << (tt[4] & 7)) & 0x80) >> 4) |
              (((data[tt[5] >> 3] << (tt[5] & 7)) & 0x80) >> 5) |
              (((data[tt[6] >> 3] << (tt[6] & 7)) & 0x80) >> 6) |
              (((data[tt[7] >> 3] << (tt[7] & 7)) & 0x80) >> 7))

    return x

  def __crypt(self, block, rounds):
    '''Encrypts or decrypts a message.
    '''
    # initial permutation
    block = self.__permutate(DES.__IP, block)
    l, r = block[:4], block[4:]
    for n in rounds:
      # expand input from 32 bits to 48 bits
      b = [x ^ y for x, y in zip(self.__kn[n], self.__permutate(DES.__E, r))]
      # define 6-bit blocks
      b = [(((b[0] & 0x80) >> 2) | ((b[0] & 0x04) << 2) |
            ((b[0] & 0x78) >> 3)),
           (((b[0] & 0x02) << 4) | (b[1] & 0x10) | ((b[0] & 0x01) << 3) |
            ((b[1] & 0xe0) >> 5)),
           (((b[1] & 0x08) << 2) | ((b[2] & 0x40) >> 2) |
            ((b[1] & 0x07) << 1) | ((b[2] & 0x80) >> 7)),
           ((b[2] & 0x20) | ((b[2] & 0x01) << 4) | ((b[2] & 0x1e) >> 1)),
           (((b[3] & 0x80) >> 2) | ((b[3] & 0x04) << 2) |
            ((b[3] & 0x78) >> 3)),
           (((b[3] & 0x02) << 4) | (b[4] & 0x10) | ((b[3] & 0x01) << 3) |
            ((b[4] & 0xe0) >> 5)),
           (((b[4] & 0x08) << 2) | ((b[5] & 0x40) >> 2) |
            ((b[4] & 0x07) << 1) | ((b[5] & 0x80) >> 7)),
           ((b[5] & 0x20) | ((b[5] & 0x01) << 4) | ((b[5] & 0x1e) >> 1))]
      # consolidate a single 32-bit block from 4-bit blocks
      s = [(DES.__S[0][b[0]] << 4) | DES.__S[1][b[1]],
           (DES.__S[2][b[2]] << 4) | DES.__S[3][b[3]],
           (DES.__S[4][b[4]] << 4) | DES.__S[5][b[5]],
           (DES.__S[6][b[6]] << 4) | DES.__S[7][b[7]]]
      # permutate the result of the iteration
      l, r = r, [x ^ y for x, y in zip(l, self.__permutate(DES.__P, s))]
    # final permutation (the inverse of the initial one)
    return self.__permutate(DES.__IP_1, r + l)

  def copy(self):
    '''Makes a separate copy of this cipher object.
    '''
    other = self.__class__(self.key)
    return other

  def encrypt(self, message):
    '''Encrypts a message.
    '''
    if len(message) % self.block_size != 0:
      raise ValueError('Cipher message length must be multiple of 8 bytes')

    result = ''
    rounds = range(0, 16, 1)
    # process data in 8-byte chunks
    for i in range(0, len(message), self.block_size):
      block = [ord(c) for c in message[i:i + self.block_size]]
      if self.__mode == 'cbc':
        block = [x ^ y for x, y in zip(self.__iv, block)]
      block = self.__crypt(block, rounds)
      if self.__mode == 'cbc':
        self.__iv = block
      result = ''.join([result, ''.join(chr(c) for c in block)])
    return result

  def decrypt(self, message):
    '''Decrypts a message.
    '''
    if len(message) % self.block_size != 0:
      raise ValueError('Cipher message length must be multiple of 8 bytes')

    result = ''
    rounds = range(15, -1, -1)
    # process data in 8-byte chunks
    for i in range(0, len(message), self.block_size):
      cblock = [ord(c) for c in message[i:i + self.block_size]]
      block = self.__crypt(cblock, rounds)
      if self.__mode == 'cbc':
        block = [x ^ y for x, y in zip(self.__iv, block)]
        self.__iv = cblock
      result = ''.join([result, ''.join(chr(c) for c in block)])
    return result

  @property
  def name(self):
    '''Name of the cipher algorithm.
    '''
    return 'des-' + mode

  @property
  def key(self):
    '''Current key of the cipher.
    '''
    return ''.join([chr(b) for b in self.__key])

  @property
  def block_size(self):
    '''Current block size of the cipher in bytes.
    '''
    return 8

  @property
  def key_size(self):
    '''Current key size of the cipher in bytes.
    '''
    return 8

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# Registry secrets
#
# These Windows registry decoder functions handle:
# * user password LM/NT hashes
# * lsa secrets
# * cached domain credentials
# * protected storage system provider
#
# Functions related to user password LM and NT hashes, LSA secrets and cached
# domain credentials are based upon Brendan Dolan-Gavitt's code in the
# 'creddump'[1] distribution and his articles 'SysKey and the SAM'[2],
# 'Decrypting LSA Secrets'[3] and 'Cached Domain Credentials'[4].
#
# Some minor extensions have been made in order to handle user password hashes
# and cached domain credentials for Windows NT 4, and some more data fields
# have been interpreted in cached domain credentials.
#
# LSA secrets functions can handle only Windows 2000/XP versions, neither
# Windows NT 4 nor Windows Vista/7 secrets are currently decrypted. Therefore,
# cached domain user hashes cannot be read in Windows NT 4 and cached
# credentials cannot be read at all in Windows Vista/7.
#
# Moreover, it has been implemented a decoder for reading the Protected Storage
# System Provider. Even though the meaning of some fields are still not known,
# data can be correctly decrypted and recovered in most part of the cases.
#
# References:
# [1] http://code.google.com/p/creddump/ and
#     http://moyix.blogspot.com/2008/02/creddump-extract-credentials-from.html
# [2] http://moyix.blogspot.com/2008/02/syskey-and-sam.html
# [3] http://moyix.blogspot.com/2008/02/decrypting-lsa-secrets.html
# [4] http://moyix.blogspot.com/2008/02/cached-domain-credentials.html
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

# user password LM/NT hashes -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

def get_syskey(reg):
  '''System bootkey.
  '''
  try:
    # windows 2k and above
    path = 'HKLM/SYSTEM/CurrentControlSet/Control/Lsa/'

    syskey = binascii.a2b_hex(''.join(reg.get_key(path + i).classname \
                              for i in ('JD', 'Skew1', 'GBG', 'Data')))

    # permutation matrix for boot key
    return ''.join(syskey[i] for i in ( 8,  5,  4,  2, 11,  9, 13,  3,
                                        0,  6,  1, 12, 14, 10, 15,  7))

  except Exception as e:
    # windows nt
    return None

def get_hashed_syskey(reg):
  '''Hashed system bootkey.
  '''
  try:
    # windows 2k and above
    f = reg.get_data('HKLM/SAM/SAM/Domains/Account/F')

    md5 = hashlib.md5(f[112:128] +
                      '!@#$%^&*()qwertyUIOPAzxcvbnmQQQQQQQQQQQQ)(*@&%\0' +
                      get_syskey(reg) +
                      '0123456789012345678901234567890123456789\0')
    rc4 = RC4(md5.digest())

    return rc4.encrypt(f[128:160])

  except Exception as e:
    # windows nt
    return None

def rid_to_key(rid):
  '''Converts the user RID into two cryptographic keys.
  '''
  deskey = ''.join([chr(rid & 0xff), chr((rid >> 8) & 0xff),
                    chr((rid >> 16) & 0xff), chr((rid >> 24) & 0xff)])

  key1 = deskey + deskey[:3]
  key2 = deskey[3] + deskey + deskey[:2]

  return key1, key2

def decrypt_user_hash(hashed_syskey, rid, encoded_lmnthash, lmntstr):
  '''Decrypts the encoded password LM/NT hash.

  lmntstr: 'LMPASSWORD\x00' for LM hash or 'NTPASSWORD\x00' for NT hash.
  '''
  try:
    # windows 2k and above
    md5 = hashlib.md5(hashed_syskey[:16] + struct.pack('<I', rid) + lmntstr)
    rc4 = RC4(md5.digest())
    obfkey = rc4.decrypt(encoded_lmnthash)

  except Exception as e:
    # windows nt
    obfkey = encoded_lmnthash

  key1, key2 = rid_to_key(rid)
  des1, des2 = DES(key1), DES(key2)
  lmnthash = des1.decrypt(obfkey[:8]) + des2.decrypt(obfkey[8:])

  return lmnthash

def encrypt_user_hash(hashed_syskey, rid, lmnthash, lmntstr):
  '''Encrypts the password LM/NT hash.

  lmntstr: 'LMPASSWORD\x00' for LM hash or 'NTPASSWORD\x00' for NT hash.
  '''
  key1, key2 = rid_to_key(rid)
  des1, des2 = DES(key1), DES(key2)
  obfkey = des1.encrypt(lmnthash[:8]) + des2.encrypt(lmnthash[8:])

  try:
    # windows 2k and above
    md5 = hashlib.md5(hashed_syskey[:16] + struct.pack('<I', rid) + lmntstr)
    rc4 = RC4(md5.digest())
    encoded_lmnthash = rc4.encrypt(obfkey)

  except Exception as e:
    # windows nt
    encoded_lmnthash = obfkey

  return encoded_lmnthash

def read_user_hashes(reg, rid):
  '''Reads the user password hashes given the RID.
  '''
  def get_v_data(v, begin, end):
    '''Unpacks user v data field.
    '''
    offset, size = struct.unpack('<II', v[begin:end])
    return v[0xcc + offset:0xcc + offset + size]

  try:
    hashed_syskey = get_hashed_syskey(reg)
    v = reg.get_data('HKLM/SAM/SAM/Domains/Account/Users/%08X/V' % rid)

    username = get_v_data(v, 12, 20).decode('utf-16-le')
    fullname = get_v_data(v, 24, 32).decode('utf-16-le')
    comment = get_v_data(v, 36, 44).decode('utf-16-le')
    usercomment = get_v_data(v, 48, 56).decode('utf-16-le')
    homedir = get_v_data(v, 72, 80).decode('utf-16-le')

    encoded_lmhash = get_v_data(v, 156, 164)
    sz = 4 if len(encoded_lmhash) != 16 else 0
    encoded_lmhash = encoded_lmhash[sz:]
    lmhash = decrypt_user_hash(hashed_syskey, rid, encoded_lmhash,
                               'LMPASSWORD\x00')

    encoded_nthash = get_v_data(v, 168, 176)
    sz = 4 if len(encoded_nthash) != 16 else 0
    encoded_nthash = encoded_nthash[sz:]
    nthash = decrypt_user_hash(hashed_syskey, rid, encoded_nthash,
                               'NTPASSWORD\x00')

    return (rid, username, lmhash, nthash, fullname, comment, usercomment,
            homedir, encoded_lmhash, encoded_nthash)

  except Exception as e:
    return None

def list_users_hashes(reg):
  '''Lists the user password LM/NT hashes.
  '''
  users = reg.get_key('HKLM/SAM/SAM/Domains/Account/Users') or []

  for user in [key for key in users if key.name != 'Names']:
    result = read_user_hashes(reg, int(user.name, 16))

    if result is not None:
      yield result

def generate_half_lmhash(password):
  '''Generates the password half LM hash.
  '''
  key = password[:7].upper()
  key = key + ('\x00' * (7 - len(key)))
  des = DES(key)
  return des.encrypt('KGS!@#$%')

def generate_lmhash(password):
  '''Generates the password LM hash.
  '''
  key = password[:14].upper()
  key = key + ('\x00' * (14 - len(key)))
  return generate_half_lmhash(key[:7]) + generate_half_lmhash(key[7:])

def generate_nthash(password):
  '''Generates the password NT hash.
  '''
  md4 = hashlib.new('md4')
  md4.update(password.encode('utf-16-le'))
  return md4.digest()

# lsa secrets =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

def get_lsakey(reg):
  '''LSA (local security authority) key.
  '''
  try:
    # windows 2k/xp
    polseckey = reg.get_data('HKLM/SECURITY/Policy/PolSecretEncryptionKey/(default)')
    md5 = hashlib.md5(get_syskey(reg) + (polseckey[60:76] * 1000))
    rc4 = RC4(md5.digest())
    return rc4.decrypt(polseckey[12:60])[16:32]

  except Exception as e:
   # windows nt/vista/7
   return None

def decrypt_sysfcn5(key, ciphertext):
  '''Decrypts data accordingly to SystemFunction005.
  '''
  plaintext, j = '', 0

  for i in range(0, len(ciphertext), 8):
    des = DES(key[j:j + 7])
    plaintext = ''.join([plaintext, des.decrypt(ciphertext[i:i + 8])])
    j = j + 7
    if len(key) - j < 7:
      j = len(key) - j

  return plaintext

def decrypt_lsa_secret(lsakey, value):
  '''Decrypts LSA secret data.
  '''
  try:
    # windows 2k/xp
    secret = decrypt_sysfcn5(lsakey, value[12:])
    size = struct.unpack('<I', secret[:4])[0]
    return secret[8:8 + size]

  except Exception as e:
    # windows nt/vista/7
    return None

def read_lsa_secret(reg, name):
  '''Reads the LSA secret given its name.
  '''
  try:
    lsakey = get_lsakey(reg)
    key = reg.get_key('HKLM/SECURITY/Policy/Secrets/%s' % name)

    currval = key.get_key('CurrVal').get_data('(default)')
    if currval is not None:
      currval = decrypt_lsa_secret(lsakey, currval)

    oldval = key.get_key('OldVal').get_data('(default)')
    if oldval is not None:
      oldval = decrypt_lsa_secret(lsakey, oldval)

    if currval is None and oldval is None:
      return None

    cupdtime = key.get_key('CupdTime').get_data('(default)')
    cupdtime = struct.unpack('<Q', cupdtime[:8])[0]
    cupdtime = datetime.datetime(1600, 1, 1, 0, 0, 0) + \
               datetime.timedelta(seconds=cupdtime // 10000000)

    oupdtime = key.get_key('OupdTime').get_data('(default)')
    oupdtime = struct.unpack('<Q', oupdtime[:8])[0]
    oupdtime = datetime.datetime(1600, 1, 1, 0, 0, 0) + \
               datetime.timedelta(seconds=oupdtime // 10000000)

    return (name, currval, oldval, cupdtime, oupdtime)

  except Exception as e:
    return None

def list_lsa_secrets(reg):
  '''Lists the LSA secrets.
  '''
  for key in reg.get_key('HKLM/SECURITY/Policy/Secrets') or []:
    result = read_lsa_secret(reg, key.name)

    if result is not None:
      yield result

def lsa_to_string(name, value):
  '''Converts a LSA secret value into a printable format.

  Known LSA secrets:
  * $MACHINE.ACC: computer account password on the domain controller.
  * NL$KM: cached domain password hashes encryption key.
  * L$_RasDefaultCredentials#0: default dial-up/VNC parameters.
  * L$HYDRAENCKEY_28ada6da-d622-11d1-9cb9-00c04fb16e75: public RSA2 key used
    in the Remote Desktop Protocol.
  * L$RTMTIMEBOMB_1320153D-8DA3-4e8e-B27B-0D888223A588: limit date to activate
    the Windows copy.
  * RasDialParams!S-1-5-21-xxxxxxxxxx-xxxxxxxxxx-xxxxxxxxx-xxxx#0: user
    dial-up/VNC parameters.
  * DefaultPassword: default password used if Windows auto logon is enabled.
  * aspnet_WP_PASSWORD: 'ASPNET' user password.
  * SCM:{3D14228D-FBE1-11D0-995D-00C04FD919C1}\x00: IWAM_xxx user password.
  * 0083343a-f925-4ed7-b1d6-d95d17a0b57b-RemoteDesktopHelpAssistantAccount:
    'HelpAssistant' user password.
  * 0083343a-f925-4ed7-b1d6-d95d17a0b57b-RemoteDesktopHelpAssistantSID:
    'HelpAssistant' user SID.
  * c261dd33-c55b-4a37-924b-746bbf3569ad-RemoteDesktopHelpAssistantEncrypt:
    'HelpAssistant' user GUID.
  '''
  try:
    if name in ('0083343a-f925-4ed7-b1d6-d95d17a0b57b-RemoteDesktopHelpAssistantAccount',
                'c261dd33-c55b-4a37-924b-746bbf3569ad-RemoteDesktopHelpAssistantEncrypt',
                'SCM:{3D14228D-FBE1-11D0-995D-00C04FD919C1}\x00',
                'aspnet_WP_PASSWORD', 'DefaultPassword') or \
       name.startswith('L$_RasDefaultCredentials#') or name.startswith('RasDialParams!'):
      return '\n'.join(value.decode('utf-16-le').rstrip('\x00').split('\x00'))

    if name == '0083343a-f925-4ed7-b1d6-d95d17a0b57b-RemoteDesktopHelpAssistantSID':
      return '-'.join([str(i) for i in struct.unpack('<cB6xBIIIII', 'S' + value)])

    if name == 'L$RTMTIMEBOMB_1320153D-8DA3-4e8e-B27B-0D888223A588':
      return get_nt_datetime(struct.unpack('<Q', value[:8])[0])

    return ' '.join(['%02x' % ord(c) for c in value])

  except Exception as e:
    if value is not None:
      return ' '.join(['%02x' % ord(c) for c in value])

    return None

# cached domain credentials =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

def decrypt_cached_credential(nlkm, ch, data):
  '''Decrypts cached domain data.

  nlkm: 'NL$KM' LSA secret current value.
  '''
  try:
    # windows 2k and above
    rc4 = RC4(hmac.new(nlkm, ch).digest())
    return rc4.decrypt(data)

  except Exception as e:
    # windows nt
    return data

def read_cached_credential(reg, name):
  '''Reads the cached domain credential given a key name.
  '''
  def get_next_offset(offset, fieldlen):
    '''Calculates the next data field offset.
    '''
    return offset + fieldlen + (fieldlen & 0x02)

  try:
    item = reg.get_value('HKLM/SECURITY/Cache/%s' % name)

    (userlen, domainlen, namelen, fullnamelen, logonlen, count,
      domainlen2, domainsidlen, hashlen, fulldomainlen) = \
      struct.unpack('<HHHHH14xH2xHH24xH2xH', item.data[:62])

    if userlen == 0:
      return None

    if hashlen > 0:
      # windows 2k and above
      nlkm = read_lsa_secret(reg, 'NL$KM')[1]
      data = decrypt_cached_credential(nlkm, item.data[64:80], item.data[96:])
      mscachehash = data[:16]
      offset = 72
    else:
      # windows nt
      data = item.data[64:]
      mscachehash = ''
      offset = 0

    username = data[offset:offset + userlen].decode('utf-16-le')
    offset = get_next_offset(offset, userlen)

    domain = data[offset:offset + domainlen].decode('utf-16-le')
    offset = get_next_offset(offset, domainlen)

    fulldomain = data[offset:offset + fulldomainlen].decode('utf-16-le')
    offset = get_next_offset(offset, fulldomainlen)

    name = data[offset:offset + namelen].decode('utf-16-le')
    offset = get_next_offset(offset, namelen)

    fullname = data[offset:offset + fullnamelen].decode('utf-16-le')
    offset = get_next_offset(offset, fullnamelen)

    logon = data[offset:offset + logonlen].decode('utf-16-le')
    offset = get_next_offset(offset, logonlen) + (8 * count)

    domain2 = data[offset:offset + domainlen2].decode('utf-16-le')
    offset = get_next_offset(offset, domainlen2)

    domainsid = '-'.join([str(i) for i in struct.unpack('<cB6xBIIII',
                         'S' + data[offset:offset + domainsidlen])])
    offset = offset + domainsidlen

    computer = data[offset:].decode('utf-16-le').rstrip('\x00')

    return (username, mscachehash, domain, fulldomain, name, fullname, logon,
            domain2, domainsid, computer)

  except Exception as e:
    return None

def list_cached_credentials(reg):
  '''Lists the cached domain credentials.
  '''
  cache = reg.get_key('HKLM/SECURITY/Cache')
  cache = [] if cache is None else cache.values

  for item in [key for key in cache if key.name != 'NL$Control']:
    result = read_cached_credential(reg, item.name)

    if result is not None:
      yield result

def generate_mscachehash(password, username):
  '''Generates the password MS-Cache hash for a given user name.
  '''
  md4 = hashlib.new('md4')
  md4.update(generate_nthash(password))
  md4.update(username.lower().encode('utf-16-le'))
  return md4.digest()

# protected storage system provider =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

def unpad(message):
  '''Gets the message without padding bytes, using the PKCS #7 version 1.5
  padding scheme.
  '''
  padlen = struct.unpack('<B', message[-1])[0]
  if padlen < 1 or padlen > 8:
    return None

  for i in range(1, padlen + 1):
    if struct.unpack('<B', message[-i])[0] != padlen:
      return None

  return message[:-padlen]

def decrypt_pstore_data2(kuser):
  '''Decrypts Protected Storage System Provider 'Data 2' subkey.

  kuser: user protected storage system provider registry key,
         'HKU/<sid>/Software/Microsoft/Protected Storage System Provider/<sid>'
  '''
  try:
    keys = {}

    # cryptographic key initialization
    sidhash = hashlib.sha1('\x66\x41\xa3\x29')
    sidhash.update(kuser.name.encode('utf-16-le'))
    sidhash.update('\x14\x9a' * (len(kuser.name) % 2))

    for subkey in kuser.get_key('Data 2') or []:
      data = subkey.get_data('Value')

      # generate the cryptographic key
      datahash = hashlib.sha1(data[40:56])
      flag = struct.unpack('<I', data[8:12])[0]
      size = 20 if flag == 3 else 4 if flag == 2 else 0
      datahash.update(sidhash.digest()[:size])

      # decrypt cryptographic key in 'Data 2'
      des = DES(datahash.digest()[:8], 'cbc')
      key = unpad(des.decrypt(data[12:36]))
      if key is not None:
        keys[subkey.name] = key
      else:
        print('Warning: decryption failure: %s' % kuser.name)

    return keys

  except Exception as e:
    return {}

def decrypt_pstore(data2, behavior, itemdata):
  '''Decrypts Protected Storage System Provider item data.
  '''
  try:
    # get the cryptographic key in 'Data 2' using name in 'Behavior'
    size = struct.unpack('<I', behavior[8:12])[0]
    name = behavior[12:12 + size].decode('utf-16-le').rstrip('\x00')
    key = data2[name]

    # decrypt cryptographic key in 'Item Data'
    des = DES(key[:8], 'cbc')
    key = unpad(des.decrypt(itemdata[8:32]))

    # decrypt data in 'Item Data'
    des = DES(key[:8], 'cbc')
    size = struct.unpack('<I', itemdata[32:36])[0]
    data = unpad(des.decrypt(itemdata[36:36 + size]))
    size = struct.unpack('<I', data[:4])[0]

    return data[4:4 + size]

  except Exception as e:
    return None

def read_pstore(reg, hive, sid, typ, subtype, name):
  '''Reads Protected Storage System Provider data given a resource name.

  Protected Storage System Provider keys:
  * InfoDelivery: 14d96c20-255b-11d1-898f-00c04fb6bfc4
    - Subscription: 00000000-0000-0000-0000-000000000000
  * Identification (Outlook Express): 220d5cc1-853a-11d0-84bc-00c04fd43f8f
    - INETCOMM Server Passwords: 417e2d75-84bd-11d0-84bb-00c04fd43f8f
  * (FTP, Front Page and HTTP passwords): 5e7e8100-9138-11d1-945a-00c04fc308ff
    - 00000000-0000-0000-0000-000000000000
      . username and passwords seperated by ':'
  * IdentityMgr: 89c39569-6841-11d2-9f59-0000f8085266
    - Identities: 0afd6bc4-beb6-4efd-8458-ced7bb39ce8f
      . IdentitiesPass: a GUID?
  * MSN: b9819c52-1e12-4bbd-91e7-af9afea5b02d
    - MSN (e-mail): 87f5aab7-ca62-41c5-8896-27714d7b7e73
  * Internet Explorer: e161255a-37c3-11d2-bcaa-00c04fd929db
    - Internet Explorer (AutoComplete): e161255a-37c3-11d2-bcaa-00c04fd929db
      . Google Search History: q:StringData/q:StringIndex
  * MSN: f09b5ab7-7aae-4c90-a638-5f11a923bc76
    - MSN (string index/data): 63809be8-c8ee-01ca-0300-000051157e0f
  '''
  try:
    kuser = reg.get_key('HKU/%s/Software/Microsoft/Protected Storage System Provider/%s'
                        % (hive, sid))
    kname = kuser.get_key('Data/%s/%s/%s' % (typ, subtype, name))
    data2 = decrypt_pstore_data2(kuser)

    if kname is not None:
      data = decrypt_pstore(data2, kname.get_data('Behavior'),
                            kname.get_data('Item Data'))
      index = None

      # Internet Explorer
      if name.endswith(':StringIndex'):
        index, data = data, None

      # MSN
      if typ == 'f09b5ab7-7aae-4c90-a638-5f11a923bc76' and \
        subtype == '63809be8-c8ee-01ca-0300-000051157e0f':
        offset, indexsize, datasize = struct.unpack('<III', data[:12])
        index = data[offset:offset + indexsize]
        data = data[offset + indexsize:offset + indexsize + datasize]
    else:
      kname = kuser.get_key('Data/%s/%s/%s:StringData' % (typ, subtype, name))
      data = decrypt_pstore(data2, kname.get_data('Behavior'),
                            kname.get_data('Item Data'))

      kname = kuser.get_key('Data/%s/%s/%s:StringIndex' % (typ, subtype, name))
      index = decrypt_pstore(data2, kname.get_data('Behavior'),
                             kname.get_data('Item Data'))

    return (data, index, name, kuser.get_key('Data/%s' %
            typ).get_data('Display String') or typ, kuser.get_key('Data/%s/%s' %
            (typ, subtype)).get_data('Display String') or subtype,
            get_username_from_sid(reg, sid) or sid,
            get_username_from_sid(reg, hive) or hive)

  except Exception as e:
    return None

def list_pstore(reg):
  '''Lists the Protected Storage System Provider data.
  '''
  for khive in reg.get_key('HKU') or []:
    for kuser in khive.get_key('Software/Microsoft/Protected Storage System Provider') or []:
      for ktype in kuser.get_key('Data') or []:
        for ksubtype in ktype:
          values = set()
          for kname in ksubtype:
            name = kname.name

            if name.endswith(':StringData'):
              name = name[:-11]
            elif name.endswith(':StringIndex'):
              name = name[:-12]

            if name not in values:
              result = read_pstore(reg, khive.name, kuser.name, ktype.name,
                                   ksubtype.name, name)

              if result is not None:
                values.add(name)
                yield result

def decode_pstore_string(data, index):
  '''Decodes a Protected Storage System Provider indexed string value.
  '''
  try:
    if not index.startswith('WICK'):
      return None

    # :StringIndex
    idx = []
    i, count = struct.unpack('<II', index[4:12])
    for x in range(count):
      offset, date, length = struct.unpack('<IQI', index[i:i + 16])
      date = datetime.datetime(1601, 1, 1, 0, 0, 0) + \
             datetime.timedelta(seconds=date // 10000000)
      idx.append((offset, length, date))
      i = i + 16

    # :StringData
    values = []
    for offset, length, date in idx:
      x = data[offset:offset + (length * 2)].decode('utf-16-le')
      values.append((x, date))

    return values

  except Exception as e:
    return None

def pstore_to_string(data, index=None, separator='; '):
  '''Converts a Protected Storage System Provider value into a printable format.
  '''
  try:
    if index is not None and data is not None:
      return separator.join('%s (%s)' % (i[0], i[1]) for i in
                            decode_pstore_string(data, index))

    if index is not None:
      return ' '.join('%02x' % ord(c) for c in index)

    try:
      return data.decode('utf-8').rstrip('\x00')
    except Exception as e:
      return ' '.join('%02x' % ord(c) for c in data)

  except Exception as e:
    return None


# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# @brief Report: os.info
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
class OSInfoReport (object):

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief initialize object
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def __init__ (self):
    self.id = 'os.info'
    self.name = 'general info'
    self.group = 'os'

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief generate report
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def generate (self, registry):
    self.viewer.clear_values ()

    # set report name
    computername = registry.get_data ('HKLM/SYSTEM/CurrentControlSet/Control/ComputerName/ComputerName/ComputerName') or ''
    self.viewer.set_report_name ('O.S. general info of computer <%s>' % computername)

    # product data
    current = 'HKLM/SOFTWARE/Microsoft/Windows NT/CurrentVersion'
    self.viewer.set_value ('product_name', registry.get_data ('%s/ProductName' % current))
    self.viewer.set_value ('csd_version', registry.get_data ('%s/CSDVersion' % current))
    self.viewer.set_value ('build', registry.get_data ('%s/BuildLab' % current))

    timestamp = registry.get_data ('%s/InstallDate' % current)
    self.viewer.set_value ('install_date', get_unix_datetime (timestamp))
    value = read_lsa_secret(registry, 'L$RTMTIMEBOMB_1320153D-8DA3-4e8e-B27B-0D888223A588')
    if value is not None:
      self.viewer.set_value ('max_activation_date', lsa_to_string(value[0], value[1]))

    self.viewer.set_value ('product_id', registry.get_data ('%s/ProductId' % current))

    # product key
    key = registry.get_key ('HKLM/SOFTWARE/Microsoft/Windows NT/CurrentVersion')
    if key is not None:
      product_name = key.get_data ('ProductName')
      digital_product_id = key.get_data ('DigitalProductId')
      product_key = get_product_key (digital_product_id)
      self.viewer.set_value ('product_key', product_key)
    
    self.viewer.set_value ('registered_owner', registry.get_data ('%s/RegisteredOwner' % current))
    self.viewer.set_value ('registered_organization', registry.get_data ('%s/RegisteredOrganization' % current))

    # winlogon
    winlogon = 'HKLM/SOFTWARE/Microsoft/Windows NT/CurrentVersion/Winlogon'
    self.viewer.set_value ('default_user', registry.get_data ('%s/DefaultUserName' % winlogon))
    self.viewer.set_value ('default_domain', registry.get_data ('%s/DefaultDomainName' % winlogon))

    # last shutdown
    timestamp = registry.get_data ('HKLM/SYSTEM/CurrentControlSet/Control/Windows/ShutdownTime')
    if timestamp is not None:
      timestamp = struct.unpack ('<Q', timestamp)[0]
      self.viewer.set_value ('last_shutdown', get_nt_datetime (timestamp))

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief build viewer
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def build_viewer (self):
    self.viewer = gdata.mediator.call ('ui.new-widget', 'attribute-list')
    self.viewer.set_report_id (self.id)
    self.viewer.set_report_app ('%s v%s' % (EXTENSION_NAME, EXTENSION_VERSION))
    self.viewer.add_value ('product_name', 'product name')
    self.viewer.add_value ('csd_version', 'CSD Version')
    self.viewer.add_value ('build', 'build')
    self.viewer.add_value ('install_date', 'install date')
    self.viewer.add_value ('max_activation_date', 'max activation date')
    self.viewer.add_value ('product_id', 'product id')
    self.viewer.add_value ('product_key', 'product key')
    self.viewer.add_value ('registered_owner', 'registered owner')
    self.viewer.add_value ('registered_organization', 'registered organization')
    self.viewer.add_value ('default_user', 'default user name')
    self.viewer.add_value ('default_domain', 'default domain name')
    self.viewer.add_value ('last_shutdown', 'last shutdown date/time')

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# @brief Report: os.folders
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
class OSFoldersReport (object):

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief initialize object
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def __init__ (self):
    self.id = 'os.folders'
    self.name = 'folders'
    self.group = 'os'

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief generate report
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def generate (self, registry):
    self.viewer.clear ()
    environment = {}

    # set report name
    computername = registry.get_data ('HKLM/SYSTEM/CurrentControlSet/Control/ComputerName/ComputerName/ComputerName') or ''
    self.viewer.set_report_name ('O.S. folders of computer <%s>' % computername)

    # bootdir
    bootdir = registry.get_data ('HKLM/SOFTWARE/Microsoft/Windows/CurrentVersion/Setup/BootDir')
    self.viewer.add_row (('BootDir', bootdir))

    # systemdrive, systemroot
    systemroot = registry.get_data ('HKLM/SOFTWARE/Microsoft/Windows NT/CurrentVersion/SystemRoot')
    systemdrive = None
    if systemroot is not None:
      systemdrive = systemroot.split ('\\')[0]
    self.viewer.add_row (('SystemDrive', systemdrive))
    self.viewer.add_row (('SystemRoot', systemroot))

    environment['systemdrive'] = systemdrive    
    environment['systemroot'] = systemroot
    
    # programdir
    programfilesdir = registry.get_data ('HKLM/SOFTWARE/Microsoft/Windows/CurrentVersion/ProgramFilesDir')
    self.viewer.add_row (('ProgramFiles', programfilesdir))

    # commonfilesdir
    commonfilesdir = registry.get_data ('HKLM/SOFTWARE/Microsoft/Windows/CurrentVersion/CommonFilesDir')
    self.viewer.add_row (('CommonFiles', commonfilesdir))

    # profilesdir
    profilesdir = registry.get_data ('HKLM/SOFTWARE/Microsoft/Windows NT/CurrentVersion/ProfileList/ProfilesDirectory')
    self.viewer.add_row (('ProfilesDir', expand_sz (profilesdir, environment)))
    
    # mediapath
    mediapath = registry.get_data ('HKLM/SOFTWARE/Microsoft/Windows/CurrentVersion/MediaPath')
    if not mediapath:
      mediapathunexpanded = registry.get_data ('HKLM/SOFTWARE/Microsoft/Windows/CurrentVersion/MediaPathUnexpanded')
      mediapath = expand_sz (mediapathunexpanded, environment)

    self.viewer.add_row (('MediaPath', mediapath))
    
    # shell folders
    key = registry.get_key ('HKLM/SOFTWARE/Microsoft/Windows/CurrentVersion/Explorer/Shell Folders')
    if key:
      for value in key.values:
        self.viewer.add_row ((value.name, value.data))
        
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief build viewer
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def build_viewer (self):
    self.viewer = gdata.mediator.call ('ui.new-widget', 'data-table')

    column = self.viewer.add_column ('name')
    column.is_sortable = True

    column = self.viewer.add_column ('path')
    column.is_sortable = True

    self.viewer.set_report_id (self.id)
    self.viewer.set_report_app ('%s v%s' % (EXTENSION_NAME, EXTENSION_VERSION))

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# @brief Report: os.tcpip-interfaces
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
class OSTCPIPInterfacesReport (object):

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief initialize object
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def __init__ (self):
    self.id = 'os.tcpip-interfaces'
    self.name = 'TCP/IP interfaces'
    self.group = 'os'

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief generate report
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def generate (self, registry):
    self.viewer.clear ()

    # set report name
    computername = registry.get_data ('HKLM/SYSTEM/CurrentControlSet/Control/ComputerName/ComputerName/ComputerName') or ''
    self.viewer.set_report_name ('TCP/IP interfaces of computer <%s>' % computername)

    # fill data
    for key in registry.get_key ('HKLM/SYSTEM/CurrentControlSet/Services/Tcpip/Parameters/Interfaces') or []:
      guid = key.name

      # get name
      connection_key = registry.get_key ('HKLM/SYSTEM/CurrentControlSet/Control/Network/{4D36E972-E325-11CE-BFC1-08002BE10318}/%s/Connection' % guid)
      if connection_key:
        name = connection_key.get_data ('Name')
      else:
        name = None

      # check DHCP
      dhcp = key.get_data ('EnableDHCP') == 1

      if dhcp:
        ip_address = key.get_data ('DhcpIPAddress')
        default_gateway = key.get_data ('DhcpDefaultGateway')
        nameservers = (key.get_data ('DhcpNameServer') or '').replace (',', '\n').replace (' ', '\n')
        dhcp_server = key.get_data ('DhcpServer')
      else:
        ip_address = key.get_data ('IPAddress')
        default_gateway = key.get_data ('DefaultGateway')
        nameservers = (key.get_data ('NameServer') or '').replace (',', '\n').replace (' ', '\n')
        dhcp_server = ''

      if name:      
        self.viewer.add_row ((name, dhcp, ip_address, default_gateway, nameservers, dhcp_server))

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief build viewer
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def build_viewer (self):
    self.viewer = gdata.mediator.call ('ui.new-widget', 'data-table')
    self.viewer.add_column ('name')
    self.viewer.add_column ('DHCP')

    column = self.viewer.add_column ('IP Address')
    column.is_sortable = True

    self.viewer.add_column ('default gateway')
    self.viewer.add_column ('nameservers')
    self.viewer.add_column ('DHCP server')

    self.viewer.set_report_id (self.id)
    self.viewer.set_report_app ('%s v%s' % (EXTENSION_NAME, EXTENSION_VERSION))

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# @brief Report: computer.info
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
class ComputerInfoReport (object):

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief initialize object
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def __init__ (self):
    self.id = 'computer.info'
    self.name = 'general info'
    self.group = 'computer'

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief generate report
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def generate (self, registry):
    self.viewer.clear_values ()

    # set report name
    computer_name = registry.get_data ('HKLM/SYSTEM/CurrentControlSet/Control/ComputerName/ComputerName/ComputerName') or ''
    self.viewer.set_report_name ('General info of computer <%s>' % computer_name)
    self.viewer.set_value ('computer_name', computer_name)

    # machine SID
    v = registry.get_data ('HKLM/SAM/SAM/Domains/Account/V')
    if v:
      machine_sid = 'S-1-5-21-%d-%d-%d' % struct.unpack ('<III', v[-12:])
    else:
      machine_sid = ''
      
    self.viewer.set_value ('machine_sid', machine_sid)

    # machine GUID
    machine_guid = registry.get_data ('HKLM/SOFTWARE/Microsoft/Cryptography/MachineGuid')
    self.viewer.set_value ('machine_guid', machine_guid)

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief build viewer
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def build_viewer (self):
    self.viewer = gdata.mediator.call ('ui.new-widget', 'attribute-list')
    self.viewer.set_report_id (self.id)
    self.viewer.set_report_app ('%s v%s' % (EXTENSION_NAME, EXTENSION_VERSION))
    self.viewer.add_value ('computer_name', 'computer name')
    self.viewer.add_value ('machine_sid', 'machine SID')
    self.viewer.add_value ('machine_guid', 'machine GUID')

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# @brief Report: os.user-accounts
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
class UserAccountReport (object):

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief initialize object
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def __init__ (self):
    self.id = 'os.user-accounts'
    self.name = 'accounts'
    self.group = 'user'

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief generate report
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def generate (self, registry):
    self.viewer.clear ()

    # set report name
    computername = registry.get_data ('HKLM/SYSTEM/CurrentControlSet/Control/ComputerName/ComputerName/ComputerName') or ''
    self.viewer.set_report_name ('user accounts of computer <%s>' % computername)

    # fill data
    users = registry.get_key ('HKLM/SAM/SAM/Domains/Account/Users') or []
    for key in users:
      if key.name.startswith ('0'):
        rid = int (key.name, 16)

        # V struct
        v = key.get_data ('V')

        offset, size = struct.unpack ('II', v[12:20])
        username = v[0xcc + offset:0xcc + offset + size].decode ('utf-16')

        offset, size = struct.unpack ('II', v[24:32])
        fullname = v[0xcc + offset:0xcc + offset + size].decode ('utf-16')

        offset, size = struct.unpack ('II', v[36:44])
        comment = v[0xcc + offset:0xcc + offset + size].decode ('utf-16')

        offset, size = struct.unpack ('II', v[48:56])
        user_comment = v[0xcc + offset:0xcc + offset + size].decode ('utf-16')

        offset, size = struct.unpack ('II', v[72:80])
        homedir = v[0xcc + offset:0xcc + offset + size].decode ('utf-16')

        self.viewer.add_row ((rid, username, fullname, comment))

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief build viewer
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def build_viewer (self):
    self.viewer = gdata.mediator.call ('ui.new-widget', 'data-table')

    column = self.viewer.add_column ('rid', column_name='RID', column_type='int')
    column.is_sortable = True

    column = self.viewer.add_column ('username')
    column.is_sortable = True
    
    self.viewer.add_column ('fullname')
    self.viewer.add_column ('comment')

    self.viewer.set_report_id (self.id)
    self.viewer.set_report_app ('%s v%s' % (EXTENSION_NAME, EXTENSION_VERSION))

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# @brief Report: os.user-logon
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
class UserLogonReport (object):

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief initialize object
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def __init__ (self):
    self.id = 'os.user-logon'
    self.name = 'logon information'
    self.group = 'user'

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief generate report
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def generate (self, registry):
    self.viewer.clear ()

    # set report name
    computername = registry.get_data ('HKLM/SYSTEM/CurrentControlSet/Control/ComputerName/ComputerName/ComputerName') or ''
    self.viewer.set_report_name ('users logon info of computer <%s>' % computername)

    # fill data
    users = registry.get_key ('HKLM/SAM/SAM/Domains/Account/Users') or []
    for key in users:
      if key.name.startswith ('0'):
        rid = int (key.name, 16)
        f = key.get_data ('F')
        v = key.get_data ('V')

        # F struct
        last_login, last_password, expire_timestamp, last_failed, sid, b1, b2, country_code, invalid_pwd_count, logon_count = struct.unpack ('8xQ8xQQQI4xBB2xH2xHH', f[:68])
        last_login_datetime = get_nt_datetime (last_login) if last_login != 0 else ''
        last_failed_datetime = get_nt_datetime (last_failed) if last_failed != 0 else ''

        # V struct
        offset, size = struct.unpack ('II', v[12:20])
        username = v[0xcc + offset:0xcc + offset + size].decode ('utf-16')

        self.viewer.add_row ((rid, username, logon_count, last_login_datetime, last_failed_datetime))

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief build viewer
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def build_viewer (self):
    self.viewer = gdata.mediator.call ('ui.new-widget', 'data-table')

    column = self.viewer.add_column ('rid', column_name='RID', column_type='int')
    column.is_sortable = True

    column = self.viewer.add_column ('username')
    column.is_sortable = True
    
    column = self.viewer.add_column ('logons', column_type='int')
    column.is_sortable = True
    
    column = self.viewer.add_column ('last_logon', column_name='last logon (UTC)')
    column.is_sortable = True

    column = self.viewer.add_column ('last_invalid_passwd', column_name='last invalid password (UTC)')
    column.is_sortable = True

    self.viewer.set_report_id (self.id)
    self.viewer.set_report_app ('%s v%s' % (EXTENSION_NAME, EXTENSION_VERSION))

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# @brief Report: user password hashes
# @author Vladimir Santos
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
class UserPasswordReport(object):

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  # @brief initialize object
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  def __init__(self):
    self.id = 'user.passwords'
    self.name = 'passwords'
    self.group = 'user'

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  # @brief generate report
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  def generate(self, registry):
    self.viewer.clear()

    # set report name
    name = registry.get_data('HKLM/SYSTEM/CurrentControlSet/Control/'
                             'ComputerName/ComputerName/ComputerName') or ''
    self.viewer.set_report_name('user passwords of computer <%s>' % name)

    # fill data
    passwords = None
    nthashes, lmhashes = {}, {}

    for user in list_users_hashes(registry):
      rid, username, lmhash, nthash = user[:4]
      password = None
      status = 'unknown'

      # try to find the password
      if nthash != '' or lmhash != '':
        if passwords is None:
          passwords = candidate_passwords(registry)

        if nthash != '':
          if len(nthashes) == 0:
            for value in passwords:
              nthashes[generate_nthash(value)] = value

          if nthashes.has_key(nthash):
            password = nthashes[nthash]
          elif generate_nthash(username) == nthash:
            password = username
  
        elif lmhash != '':
          if len(lmhashes) == 0:
            for value in passwords:
              lmhashes[generate_lmhash(value)] = value

          if lmhashes.has_key(lmhash):
            password = lmhashes[lmhash]
          elif generate_lmhash(username) == lmhash:
            password = username

      if password is not None:
        status = 'found'
      elif lmhash.endswith('\xaa\xd3\xb4\x35\xb5\x14\x04\xee'):
        status = 'smaller than 8 bytes'
      elif lmhash != '':
        status = 'greater than 7 bytes'
      elif nthash == '':
        status = 'no hashes'

      self.viewer.add_row((rid, username, binascii.b2a_hex(lmhash),
                           binascii.b2a_hex(nthash), password, status))

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  # @brief build viewer
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  def build_viewer(self):
    self.viewer = gdata.mediator.call('ui.new-widget', 'data-table')

    column = self.viewer.add_column('rid', column_name='rid', column_type='int')
    column.is_sortable = True

    column = self.viewer.add_column('username')
    column.is_sortable = True

    column = self.viewer.add_column('lmhash', column_name='lm hash')
    column.is_sortable = True

    column = self.viewer.add_column('nthash', column_name='nt hash')
    column.is_sortable = True

    column = self.viewer.add_column('password')
    column.is_sortable = True

    column = self.viewer.add_column('status', column_name='password status')
    column.is_sortable = True

    self.viewer.set_report_id(self.id)
    self.viewer.set_report_app('%s v%s' % (EXTENSION_NAME, EXTENSION_VERSION))
    self.viewer.add_export_handler('john', 'john the ripper', 'john', self.__export_john_handler)
    self.viewer.add_export_handler('pot', 'john the ripper pot', 'pot', self.__export_johnpot_handler)

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  # @brief export data as john file
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  def __export_john_handler(self, uri, data):
    # @deprecated must use URI extension instead
    path = urllib.unquote(urlparse.urlparse(uri)[2])
    fp = open(path, 'w')

    # store rows
    for rid, username, lmhash, nthash, password, status in data.rows:
      if lmhash:
        fp.write('%s:%s\n' % (username, lmhash))
      if nthash:
        fp.write('%s-nt:$NT$%s\n' % (username, nthash))

    # close file
    fp.close()

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  # @brief export passwords as john pot file
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  def __export_johnpot_handler(self, uri, data):
    # @deprecated must use URI extension instead
    path = urllib.unquote(urlparse.urlparse(uri)[2])
    fp = open(path, 'w')

    # store rows
    for rid, username, lmhash, nthash, password, status in data.rows:
      if password is not None:
        if lmhash:
          fp.write('$LM$%s:%s\n' % (lmhash[:16], password[:7]))
          if len(password) > 7:
            fp.write('$LM$%s:%s\n' % (lmhash[16:], password[7:]))
        if nthash:
          fp.write('$NT$%s:%s\n' % (nthash, password))

    # close file
    fp.close()

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# @brief Report: lsa secrets
# @author Vladimir Santos
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
class OSLSASecretReport(object):

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  # @brief initialize object
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  def __init__(self):
    self.id = 'os.lsa-secrets'
    self.name = 'lsa secrets'
    self.group = 'os'

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  # @brief generate report
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  def generate(self, registry):
    self.viewer.clear()

    # set report name
    name = registry.get_data('HKLM/SYSTEM/CurrentControlSet/Control/'
                             'ComputerName/ComputerName/ComputerName') or ''
    self.viewer.set_report_name('lsa secrets of computer <%s>' % name)

    # fill data
    for name, currval, oldval, cupdtime, oupdtime in list_lsa_secrets(registry):
      currval = lsa_to_string(name, currval)
      oldval = lsa_to_string(name, oldval)

      self.viewer.add_row((name, currval, cupdtime, oldval, oupdtime))

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  # @brief build viewer
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  def build_viewer(self):
    self.viewer = gdata.mediator.call('ui.new-widget', 'data-table')

    column = self.viewer.add_column('name')
    column.is_sortable = True

    column = self.viewer.add_column('currval', column_name='current value')
    column.is_sortable = True

    column = self.viewer.add_column('cupdtime', column_name='current update time')
    column.is_sortable = True

    column = self.viewer.add_column('oldval', column_name='old value')
    column.is_sortable = True

    column = self.viewer.add_column('oupdtime', column_name='old update time')
    column.is_sortable = True

    self.viewer.set_report_id(self.id)
    self.viewer.set_report_app('%s v%s' % (EXTENSION_NAME, EXTENSION_VERSION))

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# @brief Report: cached domain credentials
# @author Vladimir Santos
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
class UserCachedCredentialReport(object):

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  # @brief initialize object
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  def __init__(self):
    self.id = 'user.cached-credentials'
    self.name = 'cached credentials'
    self.group = 'user'

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  # @brief generate report
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  def generate(self, registry):
    self.viewer.clear()

    # set report name
    name = registry.get_data('HKLM/SYSTEM/CurrentControlSet/Control/'
                             'ComputerName/ComputerName/ComputerName') or ''
    self.viewer.set_report_name('cached credentials of computer <%s>' % name)

    # fill data
    passwords = None

    for cache in list_cached_credentials(registry):
      username, mscachehash, domain, fulldomain = cache[:4]
      fullname, computer = cache[5], cache[9]
      password = None
      status = 'unknown'

      # try to find the password
      if mscachehash != '':
        if passwords is None:
          passwords = candidate_passwords(registry)
        for value in passwords.union([username]):
          if generate_mscachehash(value, username) == mscachehash:
            password = value
            status = 'found'
            break

      self.viewer.add_row((username, fullname, domain, fulldomain, computer,
                           binascii.b2a_hex(mscachehash), password, status))

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  # @brief build viewer
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  def build_viewer(self):
    self.viewer = gdata.mediator.call('ui.new-widget', 'data-table')

    column = self.viewer.add_column('username')
    column.is_sortable = True

    column = self.viewer.add_column('fullname')
    column.is_sortable = True

    column = self.viewer.add_column('domain', column_name='domain name')
    column.is_sortable = True

    column = self.viewer.add_column('fulldomain', column_name='full domain name')
    column.is_sortable = True

    column = self.viewer.add_column('computer', column_name='computer name')
    column.is_sortable = True

    column = self.viewer.add_column('mscachehash', column_name='ms-cache hash')
    column.is_sortable = True

    column = self.viewer.add_column('password')
    column.is_sortable = True

    column = self.viewer.add_column('status', column_name='password status')
    column.is_sortable = True

    self.viewer.set_report_id(self.id)
    self.viewer.set_report_app('%s v%s' % (EXTENSION_NAME, EXTENSION_VERSION))

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# @brief Report: protected storage system provider
# @author Vladimir Santos
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
class UserProtectedStorageReport(object):

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  # @brief initialize object
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  def __init__(self):
    self.id = 'user.protected-storage'
    self.name = 'protected storage'
    self.group = 'user'

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  # @brief generate report
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  def generate(self, registry):
    self.viewer.clear()

    # set report name
    name = registry.get_data('HKLM/SYSTEM/CurrentControlSet/Control/'
                             'ComputerName/ComputerName/ComputerName') or ''
    self.viewer.set_report_name('protected storage of computer <%s>' % name)

    # fill data
    for data, index, name, typ, subtype, user, hive in list_pstore(registry):
      value = pstore_to_string(data, index, '\n')

      self.viewer.add_row((user, name, typ, subtype, value, hive))

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  # @brief build viewer
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  def build_viewer(self):
    self.viewer = gdata.mediator.call('ui.new-widget', 'data-table')

    column = self.viewer.add_column('username')
    column.is_sortable = True

    column = self.viewer.add_column('resource')
    column.is_sortable = True

    column = self.viewer.add_column('type')
    column.is_sortable = True

    column = self.viewer.add_column('subtype')
    column.is_sortable = True

    column = self.viewer.add_column('value')
    column.is_sortable = True

    column = self.viewer.add_column('hive', column_name='hive owner')
    column.is_sortable = True

    self.viewer.set_report_id(self.id)
    self.viewer.set_report_app('%s v%s' % (EXTENSION_NAME, EXTENSION_VERSION))

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# @brief Report: os.profiles
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
class UserProfileReport (object):

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief initialize object
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def __init__ (self):
    self.id = 'os.profiles'
    self.name = 'profiles'
    self.group = 'user'

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief generate report
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def generate (self, registry):
    self.viewer.clear ()

    # set report name
    computername = registry.get_data ('HKLM/SYSTEM/CurrentControlSet/Control/ComputerName/ComputerName/ComputerName') or ''
    self.viewer.set_report_name ('user profiles of computer <%s>' % computername)

    # get %systemroot% and %systemdrive%
    systemroot = registry.get_data ('HKLM/SOFTWARE/Microsoft/Windows NT/CurrentVersion/SystemRoot')
    if systemroot:
      systemdrive = systemroot.split ('\\')[0]

    else:
      systemroot = '%systemroot%'
      systemdrive = '%systemdrive%'

    # fill data
    profilelist = registry.get_key ('HKLM/SOFTWARE/Microsoft/Windows NT/CurrentVersion/ProfileList') or []
    
    for key in profilelist:
      sid = key.name
      flags = key.get_data ('Flags')

      imagepath = key.get_data ('ProfileImagePath') or ''
      imagepath = imagepath.replace ('%systemdrive%', systemdrive)
      imagepath = imagepath.replace ('%SystemDrive%', systemdrive)
      imagepath = imagepath.replace ('%systemroot%', systemroot)
      imagepath = imagepath.replace ('%SystemRoot%', systemroot)

      self.viewer.add_row ((sid, flags, imagepath))

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief build viewer
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def build_viewer (self):
    self.viewer = gdata.mediator.call ('ui.new-widget', 'data-table')

    column = self.viewer.add_column ('sid', 'SID')
    column.is_sortable = True

    self.viewer.add_column ('flags', column_type='int')
    self.viewer.add_column ('imagepath', 'image path')
    self.viewer.set_report_id (self.id)
    self.viewer.set_report_app ('%s v%s' % (EXTENSION_NAME, EXTENSION_VERSION))

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# @brief Report: installed programs
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
class InstalledProgramReport (object):

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief initialize object
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def __init__ (self):
    self.id = 'app.installed-programs'
    self.name = 'installed programs'
    self.group = 'app'

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief generate report
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def generate (self, registry):
    self.viewer.clear ()

    # set report name
    computername = registry.get_data ('HKLM/SYSTEM/CurrentControlSet/Control/ComputerName/ComputerName/ComputerName') or ''
    self.viewer.set_report_name ('installed programs of computer <%s>' % computername)

    # fill data
    uninstall_info = registry.get_key ('HKLM/SOFTWARE/Microsoft/Windows/CurrentVersion/Uninstall') or []

    for key in uninstall_info:
      displayname = key.get_data ('DisplayName')
      
      if displayname:
        publisher = key.get_data ('Publisher')

        # version
        version = key.get_data ('DisplayVersion')
        if not version:
          major = key.get_data ('VersionMajor')
          minor = key.get_data ('VersionMinor')
          
          if major:
            version = '%s.%s' % (major, minor)
            
          else:
            version = ''

        # installdate
        installdate = key.get_data ('InstallDate')

        if installdate:
          if '-' not in installdate and len (installdate) == 8:
            installdate = installdate[:4] + '-' + installdate[4:6] + '-' + installdate[6:]

        self.viewer.add_row ((displayname, version, installdate, publisher))

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief build viewer
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def build_viewer (self):
    self.viewer = gdata.mediator.call ('ui.new-widget', 'data-table')

    column = self.viewer.add_column ('name')
    column.is_sortable = True

    self.viewer.add_column ('version')

    column = self.viewer.add_column ('installdate', column_name='install date')
    column.is_sortable = True

    column = self.viewer.add_column ('publisher')
    column.is_sortable = True

    self.viewer.set_report_id (self.id)
    self.viewer.set_report_app ('%s v%s' % (EXTENSION_NAME, EXTENSION_VERSION))

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# @brief Report: computer.<devices>
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
class DevicesReport (object):

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief initialize object
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def __init__ (self, guid, name):
    self.guid = guid
    self.id = 'computer.%s' % name.replace (' ','-')
    self.name = name
    self.group = 'computer'

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief generate report
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def generate (self, registry):
    self.viewer.clear ()

    # set report name
    computername = registry.get_data ('HKLM/SYSTEM/CurrentControlSet/Control/ComputerName/ComputerName/ComputerName') or ''
    self.viewer.set_report_name ('%s of computer <%s>' % (self.name, computername))

    # get devices from DeviceClasses
    class_key = registry.get_key ('HKLM/SYSTEM/CurrentControlSet/Control/DeviceClasses/{%s}' % self.guid)

    if not class_key:
      return  

    # iter through devices
    for key in class_key:
      data = key.name.split ('#')
      device_type = data[3]
      device_class_id = data[4]
      device_id = data[5]

      # serial
      if len (device_id) > 2 and device_id[1] != '&' and device_id[-2:] == '&0':
        device_serial = device_id.split ('&')[0]
      else:
        device_serial = ''

      enum_key = registry.get_key ('HKLM/SYSTEM/CurrentControlSet/Enum/%s/%s/%s' % (device_type, device_class_id, device_id))

      if enum_key:
        device_name = enum_key.get_data ('FriendlyName') or enum_key.get_data ('DeviceDesc').split (';')[-1]
      else:
        device_name = ''

      self.viewer.add_row ((device_type, device_name, device_serial, device_id, device_class_id))

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief build viewer
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def build_viewer (self):
    self.viewer = gdata.mediator.call ('ui.new-widget', 'data-table')

    column = self.viewer.add_column ('type')
    column.is_sortable = True

    column = self.viewer.add_column ('name')
    column.is_sortable = True

    column = self.viewer.add_column ('serial', column_name='serial number')
    column.is_sortable = True

    self.viewer.add_column ('device_id', column_name='device ID')
    self.viewer.add_column ('device_class_id', column_name='device class ID')

    self.viewer.set_report_id (self.id)
    self.viewer.set_report_app ('%s v%s' % (EXTENSION_NAME, EXTENSION_VERSION))

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# @brief Report: recent docs
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
class RecentDocsReport (object):

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief initialize object
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def __init__ (self):
    self.id = 'user.recent-docs'
    self.name = 'recent docs'
    self.group = 'user'

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief generate report
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def generate (self, registry):
    self.viewer.clear ()

    # set report name
    computername = registry.get_data ('HKLM/SYSTEM/CurrentControlSet/Control/ComputerName/ComputerName/ComputerName') or ''
    self.viewer.set_report_name ('recent docs of computer <%s>' % computername)

    # fill data
    users = registry.get_key ('HKU') or []

    for key in users:
      sid = key.name
      username = get_username_from_sid (registry, sid) or sid
 
      recent_docs_key = key.get_key ('Software/Microsoft/Windows/CurrentVersion/Explorer/RecentDocs')
      if recent_docs_key:
        data = recent_docs_key.get_data ('MRUListEx') or ''
        mrusize = len (data) // 4
        mrulist = struct.unpack ('<%di' % mrusize, data)

        for seq, idx in enumerate (mrulist):
          if idx != -1:
            doc = unicode (recent_docs_key.get_data (str (idx)) or '', 'utf-16-le')
            self.viewer.add_row ((username, mrusize - seq - 1, doc))

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief build viewer
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def build_viewer (self):
    self.viewer = gdata.mediator.call ('ui.new-widget', 'data-table')

    column = self.viewer.add_column ('username')
    column.is_sortable = True

    column = self.viewer.add_column ('sequence', column_type='int')
    column.is_sortable = True
    
    column = self.viewer.add_column ('document')
    column.is_sortable = True
    column.is_searchable = True
    
    self.viewer.set_report_id (self.id)
    self.viewer.set_report_app ('%s v%s' % (EXTENSION_NAME, EXTENSION_VERSION))

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# @brief Report: user assist
# @begin-description
#   1. under UserAssist key there are GUID keys for each type of userassist.
#   2. Windows Vista and later use GUID directory instead of path. 
# @end-description
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
class UserAssistReport (object):

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief initialize object
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def __init__ (self):
    self.id = 'user.user-assist'
    self.name = 'user assist'
    self.group = 'user'

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief generate report
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def generate (self, registry):
    self.viewer.clear ()

    # set report name
    computername = registry.get_data ('HKLM/SYSTEM/CurrentControlSet/Control/ComputerName/ComputerName/ComputerName') or ''
    self.viewer.set_report_name ('user assist of computer <%s>' % computername)

    # for each user in HKU
    for key in registry.get_key ('HKU') or []:
      sid = key.name
      username = get_username_from_sid (registry, sid) or sid
      folder_paths = {}
 
      # for each GUID key inside UserAssist
      for guid_key in key.get_key ('Software/Microsoft/Windows/CurrentVersion/Explorer/UserAssist') or []:
        guid = guid_key.name
        origin = registry.get_data ('HKCR/CLSID/%s/(default)' % guid) or guid[1:-1]

        # for each entry inside Count subkey        
        for location, session, count, date in self.list_guid_entries (guid_key):

          # translate Vista and later GUID folders
          if location.startswith ('{'):
            pos = location.find ('}')
            folder_guid = location[1:pos].lower ()
            folder_path = folder_paths.get (folder_guid)

            if not folder_path:
              folder_path = registry.get_folder_path_from_guid (sid, folder_guid)
              folder_paths[folder_guid] = folder_path

            if folder_path:
              location = folder_path + location[pos+1:]

          # add entry
          self.viewer.add_row ((username, origin, session, count, str (date), location))

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief return entries for a given GUID
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def list_guid_entries (self, guid_key):
    version = guid_key.get_data ('Version' or -1)

    count_key = guid_key.get_key ('Count')
    if count_key:
          
      for value in count_key.values:
        location = value.name.encode ('rot13').decode ('cp1252')
        data = value.data

        session = -1
        count = -1
        date = ''

        if version == 3:
          if location == 'UEME_CTLSESSION':
            unknown1, session = struct.unpack ('<II', data)

          elif len (data) == 16:
            session, count, timestamp = struct.unpack ('<IIQ', data)
            date = get_nt_datetime (timestamp)

            if date:
              count = count - 5         # counter starts at 5

            else:
              count = -1

        yield location, session, count, date
        
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief build viewer
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def build_viewer (self):
    self.viewer = gdata.mediator.call ('ui.new-widget', 'data-table')

    column = self.viewer.add_column ('username')
    column.is_sortable = True

    column = self.viewer.add_column ('category')
    column.is_sortable = True

    column = self.viewer.add_column ('session', column_type='int')
    column.is_sortable = True

    column = self.viewer.add_column ('count', column_type='int')
    column.is_sortable = True

    column = self.viewer.add_column ('timestamp')
    column.is_sortable = True

    column = self.viewer.add_column ('location')
    column.is_sortable = True

    self.viewer.set_report_id (self.id)
    self.viewer.set_report_app ('%s v%s' % (EXTENSION_NAME, EXTENSION_VERSION))

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# @brief Report: computer descriptions
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
class ComputerDescriptionsReport (object):

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief initialize object
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def __init__ (self):
    self.id = 'user.computer-descriptions'
    self.name = 'computer descriptions'
    self.group = 'user'

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief generate report
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def generate (self, registry):
    self.viewer.clear ()

    # set report name
    computername = registry.get_data ('HKLM/SYSTEM/CurrentControlSet/Control/ComputerName/ComputerName/ComputerName') or ''
    self.viewer.set_report_name ('computer descriptions of computer <%s>' % computername)

    # fill data
    users = registry.get_key ('HKU') or []

    for key in users:
      sid = key.name
      username = get_username_from_sid (registry, sid) or sid
 
      cd_key = key.get_key ('Software/Microsoft/Windows/CurrentVersion/Explorer/ComputerDescriptions')
      if cd_key:
        for value in cd_key.values:
          name = value.name
          description = value.data
          self.viewer.add_row ((username, name, description))

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief build viewer
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def build_viewer (self):
    self.viewer = gdata.mediator.call ('ui.new-widget', 'data-table')

    column = self.viewer.add_column ('username')
    column.is_sortable = True

    column = self.viewer.add_column ('name')
    column.is_sortable = True
    
    column = self.viewer.add_column ('description')
    
    self.viewer.set_report_id (self.id)
    self.viewer.set_report_app ('%s v%s' % (EXTENSION_NAME, EXTENSION_VERSION))

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# @brief Report: product keys
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
class ProductKeyReport (object):

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief initialize object
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def __init__ (self):
    self.id = 'app.product-keys'
    self.name = 'product keys'
    self.group = 'app'

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief generate report
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def generate (self, registry):
    self.viewer.clear ()

    # set report name
    computername = registry.get_data ('HKLM/SYSTEM/CurrentControlSet/Control/ComputerName/ComputerName/ComputerName') or ''
    self.viewer.set_report_name ('product keys of computer <%s>' % computername)

    # Microsoft products
    for product_name, product_key in self.get_microsoft_product_keys (registry):
      self.viewer.add_row ((product_name, product_key))

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief generate Microsoft products keys
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def get_microsoft_product_keys (self, registry):

    # Microsoft Windows
    key = registry.get_key ('HKLM/SOFTWARE/Microsoft/Windows NT/CurrentVersion')
    if key:
      product_name = key.get_data ('ProductName')
      digital_product_id = key.get_data ('DigitalProductId')
      product_key = get_product_key (digital_product_id)
      yield product_name, product_key

    # Internet Explorer
    key = registry.get_key ('HKLM/SOFTWARE/Microsoft/Internet Explorer')
    if key:
      product_name = 'Windows Internet Explorer v' + key.get_data ('Version')
      digital_product_id = key.get_key ('Registration').get_data ('DigitalProductID')
      product_key = get_product_key (digital_product_id)
      yield product_name, product_key
    
    # Microsoft Office
    for version_key in registry.get_key ('/HKLM/SOFTWARE/Microsoft/Office') or []:
      for guid_key in version_key.get_key ('Registration') or []:
        product_name = guid_key.get_data ('ProductName') or guid_key.get_data ('ConvertToEdition')
        digital_product_id = guid_key.get_data ('DigitalProductId')

        if product_name and digital_product_id:
          product_key = get_product_key (digital_product_id)
          yield product_name, product_key
        
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief build viewer
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def build_viewer (self):
    self.viewer = gdata.mediator.call ('ui.new-widget', 'data-table')

    column = self.viewer.add_column ('product')
    column.is_sortable = True

    column = self.viewer.add_column ('key')
    column.is_sortable = True

    self.viewer.set_report_id (self.id)
    self.viewer.set_report_app ('%s v%s' % (EXTENSION_NAME, EXTENSION_VERSION))

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# @brief Report: email accounts
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
class EmailAccountsReport (object):

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief initialize object
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def __init__ (self):
    self.id = 'app.email-accounts'
    self.name = 'e-mail accounts'
    self.group = 'app'

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief generate report
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def generate (self, registry):
    self.viewer.clear ()

    # set report name
    computername = registry.get_data ('HKLM/SYSTEM/CurrentControlSet/Control/ComputerName/ComputerName/ComputerName') or ''
    self.viewer.set_report_name ('e-mail accounts of computer <%s>' % computername)

    # fill data
    for key in registry.get_key ('HKU') or []:
      sid = key.name
      username = get_username_from_sid (registry, sid) or sid
 
      # outlook 98-2000 and outlook express
      for subkey in key.get_key ('Software/Microsoft/Internet Account Manager/Accounts') or []:
        account_name = subkey.get_data ('SMTP Display Name')

        if account_name:
          email_address = subkey.get_data ('SMTP Email Address')
          pwd_keyname = subkey.get_data ('POP3 Password2') or \
                        subkey.get_data ('HTTPMail Password2')
          password = None
          if pwd_keyname is not None:
            pwd_keyname = pwd_keyname[2:].decode('utf-16-le').rstrip('\x00')
            value = read_pstore(registry, sid, sid,
                                '220d5cc1-853a-11d0-84bc-00c04fd43f8f',
                                '417e2d75-84bd-11d0-84bb-00c04fd43f8f',
                                pwd_keyname)
            password = pstore_to_string(value[0])
          status = 'unknown' if password is None else 'found'

          self.viewer.add_row ((username, account_name, email_address, password, status))

      # outlook 2002-2008
      for profile_key in key.get_key ('Software/Microsoft/Windows NT/CurrentVersion/Windows Messaging Subsystem/Profiles') or []:
        for subkey in profile_key.get_key ('9375CFF0413111d3B88A00104B2A6676') or []:
          account_name_data = subkey.get_data ('Display Name')
          
          if account_name_data:
            account_name = account_name_data.decode ('utf-16-le')
            email_address = subkey.get_data ('Email').decode ('utf-16-le')
            dpapi_password = subkey.get_data ('POP3 Password')

            self.viewer.add_row ((username, account_name, email_address, None, 'unknown'))

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  # @brief build viewer
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
  def build_viewer (self):
    self.viewer = gdata.mediator.call ('ui.new-widget', 'data-table')

    column = self.viewer.add_column ('username')
    column.is_sortable = True

    column = self.viewer.add_column ('name')
    column.is_sortable = True
    
    column = self.viewer.add_column ('e-mail address')
    column.is_sortable = True

    column = self.viewer.add_column ('password')
    column.is_sortable = True

    column = self.viewer.add_column ('password status')
    column.is_sortable = True

    self.viewer.set_report_id (self.id)
    self.viewer.set_report_app ('%s v%s' % (EXTENSION_NAME, EXTENSION_VERSION))

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# @brief start function
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
def pvt_start ():

  # Operating System
  gdata.mediator.call ('registry.register-report', OSInfoReport ())
  gdata.mediator.call ('registry.register-report', OSFoldersReport ())
  gdata.mediator.call ('registry.register-report', OSTCPIPInterfacesReport ())
  gdata.mediator.call ('registry.register-report', OSLSASecretReport ())

  # User
  gdata.mediator.call ('registry.register-report', UserAccountReport ())
  gdata.mediator.call ('registry.register-report', UserLogonReport ())
  gdata.mediator.call ('registry.register-report', UserPasswordReport ())
  gdata.mediator.call ('registry.register-report', UserCachedCredentialReport ())
  gdata.mediator.call ('registry.register-report', UserProtectedStorageReport ())
  gdata.mediator.call ('registry.register-report', UserProfileReport ())

  # Application
  gdata.mediator.call ('registry.register-report', InstalledProgramReport ())
  gdata.mediator.call ('registry.register-report', RecentDocsReport ())
  gdata.mediator.call ('registry.register-report', UserAssistReport ())
  gdata.mediator.call ('registry.register-report', ProductKeyReport ())
  gdata.mediator.call ('registry.register-report', EmailAccountsReport ())
  gdata.mediator.call ('registry.register-report', ComputerDescriptionsReport ())
  
  # Computer
  gdata.mediator.call ('registry.register-report', ComputerInfoReport ())

  report = DevicesReport ('53f56307-b6bf-11d0-94f2-00a0c91efb8b', 'disk devices')
  gdata.mediator.call ('registry.register-report', report)

  report = DevicesReport ('53f56308-b6bf-11d0-94f2-00a0c91efb8b', 'optical devices')
  gdata.mediator.call ('registry.register-report', report)

  report = DevicesReport ('ad498944-762f-11d0-8dcb-00c04fc3358c', 'network adapters')
  gdata.mediator.call ('registry.register-report', report)
  
  report = DevicesReport ('6bdd1fc6-810f-11d0-bec7-08002be2092f', 'imaging devices')
  gdata.mediator.call ('registry.register-report', report)
  
  report = DevicesReport ('5b45201d-f2f2-4f3b-85bb-30ff1f953599', 'display adapters')
  gdata.mediator.call ('registry.register-report', report)
  
  report = DevicesReport ('e6f07b5f-ee97-4a90-b076-33f57bf4eaa7', 'monitor devices')
  gdata.mediator.call ('registry.register-report', report)
  
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# @brief stop function
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
def pvt_stop ():
  gdata.mediator.call ('registry.unregister-report', 'os.info')
  gdata.mediator.call ('registry.unregister-report', 'os.folders')
  gdata.mediator.call ('registry.unregister-report', 'os.tcpip-interfaces')
  gdata.mediator.call ('registry.unregister-report', 'os.user-accounts')
  gdata.mediator.call ('registry.unregister-report', 'os.user-logon')
  gdata.mediator.call ('registry.unregister-report', 'user.passwords')
  gdata.mediator.call ('registry.unregister-report', 'os.lsa-secrets')
  gdata.mediator.call ('registry.unregister-report', 'user.cached-credentials')
  gdata.mediator.call ('registry.unregister-report', 'user.protected-storage')
  gdata.mediator.call ('registry.unregister-report', 'os.profiles')
  gdata.mediator.call ('registry.unregister-report', 'user.recent-docs')
  gdata.mediator.call ('registry.unregister-report', 'user.computer-descriptions')
  gdata.mediator.call ('registry.unregister-report', 'user.user-assist')
  gdata.mediator.call ('registry.unregister-report', 'computer.info')
  gdata.mediator.call ('registry.unregister-report', 'computer.disk-devices')
  gdata.mediator.call ('registry.unregister-report', 'computer.optical-devices')
  gdata.mediator.call ('registry.unregister-report', 'computer.network-adapters')
  gdata.mediator.call ('registry.unregister-report', 'computer.imaging-devices')
  gdata.mediator.call ('registry.unregister-report', 'computer.display-adapters')
  gdata.mediator.call ('registry.unregister-report', 'computer.monitor-devices')
  gdata.mediator.call ('registry.unregister-report', 'app.product-keys')
  gdata.mediator.call ('registry.unregister-report', 'app.installed-programs')
  gdata.mediator.call ('registry.unregister-report', 'app.email-accounts')

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# API initialization
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
def pvt_start_api ():
  pass
